/*
 * GlushkovTraversal.cpp
 *
 *  Created on: 13. 3. 2014
 *      Author: Tomas Pecka
 */

#include "GlushkovTraversal.h"

#include "properties/RegExpEpsilon.h"

using namespace alib;
using namespace regexp;
using namespace std;

namespace regexp
{

const GlushkovSymbol& GlushkovTraversal::findSymbol( regexp::UnboundedRegExpSymbol const * const symbol, const std::set<GlushkovSymbol> & symbolSet )
{
    auto it = find_if( symbolSet.begin( ), symbolSet.end( ), [ symbol ]( GlushkovSymbol const& gs ) -> bool {
        return gs.getSymbolPtr( ) == symbol;
    } );

    if( it == symbolSet.end( ) )
        throw( "GlushkovTraversal - Can not find GlushkovSymbol for regexp node." );

    return * it;
}

bool GlushkovTraversal::pos( GlushkovSymbol const& symbol, regexp::UnboundedRegExp const * const & node )
{
    return pos( & node->getRegExp(), symbol.getSymbolPtr( ) );
}

set<GlushkovSymbol> GlushkovTraversal::first( regexp::UnboundedRegExp const& re )
{
    set<GlushkovSymbol> firstSet, symbolSet = getSymbols( re );

    for( auto const& s : first( & re.getRegExp() ) )
        firstSet.insert( findSymbol( s, symbolSet ) );

    return firstSet;
}

set<GlushkovSymbol> GlushkovTraversal::last( regexp::UnboundedRegExp const& re )
{
    set<GlushkovSymbol> lastSet, symbolSet = getSymbols( re );

    for( auto const& s : last( & re.getRegExp() ) )
        lastSet.insert( findSymbol( s, symbolSet ) );

    return lastSet;
}

set<GlushkovSymbol> GlushkovTraversal::follow( regexp::UnboundedRegExp const& re, GlushkovSymbol const& symbol )
{
    set<GlushkovSymbol> followSet, symbolSet = getSymbols( re );

    for( auto const& s: follow( & re.getRegExp(), symbol.getSymbolPtr( ) ) )
        followSet.insert( findSymbol( s, symbolSet ) );

    return followSet;
}

// -----------------------------------------------------------------------------

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::first( regexp::UnboundedRegExpElement const * const & node )
{
    regexp::UnboundedRegExpAlternation const * const alternation = dynamic_cast<regexp::UnboundedRegExpAlternation const * const>( node );
    regexp::UnboundedRegExpConcatenation const * const concatenation = dynamic_cast<regexp::UnboundedRegExpConcatenation const * const>( node );
    regexp::UnboundedRegExpIteration const * const iteration = dynamic_cast<regexp::UnboundedRegExpIteration const * const>( node );
    regexp::UnboundedRegExpSymbol const * const symbol = dynamic_cast<regexp::UnboundedRegExpSymbol const * const>( node );
    regexp::UnboundedRegExpEmpty const * const empty = dynamic_cast<regexp::UnboundedRegExpEmpty const * const>( node );
    regexp::UnboundedRegExpEpsilon const * const eps = dynamic_cast<regexp::UnboundedRegExpEpsilon const * const>( node );

    if( alternation )
        return first( alternation );
    else if( concatenation )
        return first( concatenation );
    else if( iteration )
        return first( iteration );
    else if( eps )
        return first( eps );
    else if( empty )
        return first( empty );
    else if( symbol )
        return first( symbol );

    throw exception::AlibException( "GlushkovTraversal::first - invalid RegExpElement node" );
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::first( regexp::UnboundedRegExpAlternation const * const & node )
{
    set<regexp::UnboundedRegExpSymbol const *> ret, tmp;

    for( auto const& element : node->getElements( ) )
    {
        tmp = first( element );
        ret.insert( tmp.begin( ), tmp.end( ) );
    }

    return ret;
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::first( regexp::UnboundedRegExpConcatenation const * const & node )
{
    set<regexp::UnboundedRegExpSymbol const *> ret, tmp;

    for( auto const& element : node->getElements( ) )
    {
        tmp = first( element );
        ret.insert( tmp.begin( ), tmp.end( ) );

        if(! regexp::properties::RegExpEpsilon::languageContainsEpsilon(*element)) // If regexp of this subtree can match epsilon, then we need to add next subtree
            break;
    }

    return ret;
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::first( regexp::UnboundedRegExpIteration const * const & node )
{
    return first( & node->getElement( ) );
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::first( regexp::UnboundedRegExpSymbol const * const & node )
{
    return set<regexp::UnboundedRegExpSymbol const *> { node };
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::first( regexp::UnboundedRegExpEpsilon const * const & /* node */ )
{
    return set<regexp::UnboundedRegExpSymbol const *>( );
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::first( regexp::UnboundedRegExpEmpty const * const &  /* node */ )
{
    return set<regexp::UnboundedRegExpSymbol const *>( );
}

// ----------------------------------------------------------------------------

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::last( regexp::UnboundedRegExpElement const * const & node )
{
    regexp::UnboundedRegExpAlternation const * const alternation = dynamic_cast<regexp::UnboundedRegExpAlternation const * const>( node );
    regexp::UnboundedRegExpConcatenation const * const concatenation = dynamic_cast<regexp::UnboundedRegExpConcatenation const * const>( node );
    regexp::UnboundedRegExpIteration const * const iteration = dynamic_cast<regexp::UnboundedRegExpIteration const * const>( node );
    regexp::UnboundedRegExpSymbol const * const symbol = dynamic_cast<regexp::UnboundedRegExpSymbol const * const>( node );
    regexp::UnboundedRegExpEmpty const * const empty = dynamic_cast<regexp::UnboundedRegExpEmpty const * const>( node );
    regexp::UnboundedRegExpEpsilon const * const eps = dynamic_cast<regexp::UnboundedRegExpEpsilon const * const>( node );

    if( symbol )
        return last( symbol );
    else if( alternation )
        return last( alternation );
    else if( concatenation )
        return last( concatenation );
    else if( iteration )
        return last( iteration );
    else if( eps )
        return last( eps );
    else if( empty )
        return last( empty );

     throw exception::AlibException( "GlushkovTraversal::last - invalid RegExpElement node" );
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::last( regexp::UnboundedRegExpAlternation const * const & node )
{
    set<regexp::UnboundedRegExpSymbol const *> ret;

    for( auto const& element : node->getElements( ) )
    {
        set<regexp::UnboundedRegExpSymbol const *> tmp = last( element );
        ret.insert( tmp.begin( ), tmp.end( ) );
    }

    return ret;
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::last( regexp::UnboundedRegExpConcatenation const * const & node )
{
    set<regexp::UnboundedRegExpSymbol const *> ret, tmp;

    for( auto it = node->getElements( ).rbegin( ); it != node->getElements( ).rend( ) ; it ++ )
    {
        tmp = last( *it );
        ret.insert( tmp.begin( ), tmp.end( ) );

        if( ! regexp::properties::RegExpEpsilon::languageContainsEpsilon(**it) )
            break;
    }

    return ret;
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::last( regexp::UnboundedRegExpIteration const * const & node )
{
    return last( & node->getElement( ) );
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::last( regexp::UnboundedRegExpSymbol const * const & node )
{
    return set<regexp::UnboundedRegExpSymbol const *> { node };
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::last( regexp::UnboundedRegExpEpsilon const * const & /* node */ )
{
    return set<regexp::UnboundedRegExpSymbol const *>( );
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::last( regexp::UnboundedRegExpEmpty const * const & /* node */ )
{
    return set<regexp::UnboundedRegExpSymbol const *>( );
}

// ----------------------------------------------------------------------------

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::follow( regexp::UnboundedRegExpElement const * const & node, regexp::UnboundedRegExpSymbol const * const & symbolptr )
{
    regexp::UnboundedRegExpAlternation const * const alternation = dynamic_cast<regexp::UnboundedRegExpAlternation const * const>( node );
    regexp::UnboundedRegExpConcatenation const * const concatenation = dynamic_cast<regexp::UnboundedRegExpConcatenation const * const>( node );
    regexp::UnboundedRegExpIteration const * const iteration = dynamic_cast<regexp::UnboundedRegExpIteration const * const>( node );
    regexp::UnboundedRegExpSymbol const * const symbol = dynamic_cast<regexp::UnboundedRegExpSymbol const * const>( node );
    regexp::UnboundedRegExpEmpty const * const empty = dynamic_cast<regexp::UnboundedRegExpEmpty const * const>( node );
    regexp::UnboundedRegExpEpsilon const * const eps = dynamic_cast<regexp::UnboundedRegExpEpsilon const * const>( node );

    if( alternation )
        return follow( alternation, symbolptr );

    else if( concatenation )
        return follow( concatenation, symbolptr );

    else if( iteration )
        return follow( iteration, symbolptr );

    else if( symbol )
        return follow( symbol, symbolptr );

    else if( empty )
        return follow( empty, symbolptr );

    else if( eps )
        return follow( eps, symbolptr );

    throw exception::AlibException( "GlushkovTraversal::follow() - unknown RegExpElement node" );
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::follow( regexp::UnboundedRegExpAlternation const * const & node, regexp::UnboundedRegExpSymbol const * const & symbolptr )
{
    for( auto const& element : node->getElements( ) )
        if( pos( element, symbolptr ) )
            return follow( element, symbolptr );

    throw exception::AlibException( "GlushkovTraversal::follow(Alt)" );
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::follow( regexp::UnboundedRegExpConcatenation const * const & node, regexp::UnboundedRegExpSymbol const * const & symbolptr )
{
    set<regexp::UnboundedRegExpSymbol const *> ret, tmp, lastSet;

    for( auto e = node->getElements( ).begin( ); e != node->getElements( ).end( ); e ++ )
    {
        if( ! pos( *e, symbolptr ) )
            continue;

        tmp = follow( *e, symbolptr );
        ret.insert( tmp.begin( ), tmp.end( ) );

        lastSet = last( *e );
        if( isInSet( symbolptr, lastSet ) )
        {
            for( auto f = next( e ); f != node->getElements( ).end( ); f ++ )
            {
                tmp = first( *f );
                ret.insert( tmp.begin( ), tmp.end( ) );

                if( ! regexp::properties::RegExpEpsilon::languageContainsEpsilon( **f ) )
                    break;
            }
        }
    }
    return ret;
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::follow( regexp::UnboundedRegExpIteration const * const & node, regexp::UnboundedRegExpSymbol const * const & symbolptr )
{
    set<regexp::UnboundedRegExpSymbol const *> ret = follow( & node->getElement( ), symbolptr );
    set<regexp::UnboundedRegExpSymbol const *> lastSet = last( & node->getElement( ) );

    if( isInSet( symbolptr, lastSet ) )
    {
        set<regexp::UnboundedRegExpSymbol const *> firstSet = first( & node->getElement( ) );
        ret.insert( firstSet.begin( ), firstSet.end( ) );
    }

    return ret;
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::follow( regexp::UnboundedRegExpSymbol const * const & /* node */, regexp::UnboundedRegExpSymbol const * const & /* symbolptr */ )
{
    return set<regexp::UnboundedRegExpSymbol const *>( );
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::follow( regexp::UnboundedRegExpEmpty const * const & /* node */, regexp::UnboundedRegExpSymbol const * const & /* symbolptr */ )
{
    return set<regexp::UnboundedRegExpSymbol const *>( );
}

set<regexp::UnboundedRegExpSymbol const *> GlushkovTraversal::follow( regexp::UnboundedRegExpEpsilon const * const & /* node */, regexp::UnboundedRegExpSymbol const * const & /* symbolptr */ )
{
    return set<regexp::UnboundedRegExpSymbol const *>( );
}

// ----------------------------------------------------------------------------

bool GlushkovTraversal::pos( regexp::UnboundedRegExpElement const * const & node, regexp::UnboundedRegExpSymbol const * const & symbolptr )
{
    regexp::UnboundedRegExpAlternation const * const alternation = dynamic_cast<regexp::UnboundedRegExpAlternation const * const>( node );
    regexp::UnboundedRegExpConcatenation const * const concatenation = dynamic_cast<regexp::UnboundedRegExpConcatenation const * const>( node );
    regexp::UnboundedRegExpIteration const * const iteration = dynamic_cast<regexp::UnboundedRegExpIteration const * const>( node );
    regexp::UnboundedRegExpSymbol const * const symbol = dynamic_cast<regexp::UnboundedRegExpSymbol const * const>( node );
    regexp::UnboundedRegExpEmpty const * const empty = dynamic_cast<regexp::UnboundedRegExpEmpty const * const>( node );
    regexp::UnboundedRegExpEpsilon const * const eps = dynamic_cast<regexp::UnboundedRegExpEpsilon const * const>( node );

    if( alternation )
        return pos( alternation, symbolptr );

    else if( concatenation )
        return pos( concatenation, symbolptr );

    else if( iteration )
        return pos( iteration, symbolptr );

    else if( symbol )
        return pos( symbol, symbolptr );

    else if( empty )
        return pos( empty, symbolptr );

    else if( eps )
        return pos( eps, symbolptr );

    throw exception::AlibException( "GlushkovTraversal::pos() - unknown RegExpElement node" );
}

bool GlushkovTraversal::pos( regexp::UnboundedRegExpAlternation const * const & node, regexp::UnboundedRegExpSymbol const * const & symbolptr )
{
    for( auto const& element : node->getElements( ) )
        if( pos( element, symbolptr ) )
            return true;
    return false;
}

bool GlushkovTraversal::pos( regexp::UnboundedRegExpConcatenation const * const & node, regexp::UnboundedRegExpSymbol const * const & symbolptr )
{
    for( auto const& element : node->getElements( ) )
        if( pos( element, symbolptr ) )
            return true;
    return false;
}

bool GlushkovTraversal::pos( regexp::UnboundedRegExpIteration const * const & node, regexp::UnboundedRegExpSymbol const * const & symbolptr )
{
    return pos( & node->getElement( ), symbolptr );
}

bool GlushkovTraversal::pos( regexp::UnboundedRegExpSymbol const * const & node, regexp::UnboundedRegExpSymbol const * const & symbolptr )
{
    return symbolptr == node;
}

bool GlushkovTraversal::pos( regexp::UnboundedRegExpEmpty const * const & /* node */, regexp::UnboundedRegExpSymbol const * const & /* symbolptr */ )
{
    return false;
}

bool GlushkovTraversal::pos( regexp::UnboundedRegExpEpsilon const * const & /* node */, regexp::UnboundedRegExpSymbol const * const & /* symbolptr */ )
{
    return false;
}

// ----------------------------------------------------------------------------

set<GlushkovSymbol> GlushkovTraversal::getSymbols( regexp::UnboundedRegExp const& re )
{
    set<GlushkovSymbol> alphabet;
    int i = 1;

    getSymbols( & re.getRegExp( ), alphabet, i );

    return alphabet;
}

void GlushkovTraversal::getSymbols( regexp::UnboundedRegExpElement const * const & node, set<GlushkovSymbol> & alphabet, int & i )
{
    regexp::UnboundedRegExpAlternation const * const alternation = dynamic_cast<regexp::UnboundedRegExpAlternation const * const>( node );
    regexp::UnboundedRegExpConcatenation const * const concatenation = dynamic_cast<regexp::UnboundedRegExpConcatenation const * const>( node );
    regexp::UnboundedRegExpIteration const * const iteration = dynamic_cast<regexp::UnboundedRegExpIteration const * const>( node );
    regexp::UnboundedRegExpSymbol const * const symbol = dynamic_cast<regexp::UnboundedRegExpSymbol const * const>( node );
    regexp::UnboundedRegExpEmpty const * const empty = dynamic_cast<regexp::UnboundedRegExpEmpty const * const>( node );
    regexp::UnboundedRegExpEpsilon const * const eps = dynamic_cast<regexp::UnboundedRegExpEpsilon const * const>( node );

    if( symbol )
    {
        alphabet.insert( GlushkovSymbol( symbol, i ++ ) );
        return;
    }

    else if( alternation )
    {
        for( const auto & element : alternation->getElements( ) )
            getSymbols( element, alphabet, i );
        return;
    }

    else if( concatenation )
    {
        for( const auto & element : concatenation->getElements( ) )
            getSymbols( element, alphabet, i );
        return;
    }

    else if( iteration )
    {
        getSymbols( & iteration->getElement( ), alphabet, i );
        return;
    }

    else if( empty )
        return;

    else if( eps )
        return;

    throw exception::AlibException( "GlushkovTraversal::getSymbols() - unknown RegExpElement node" );
}

} /* namespace conversions */
