/*
 * RegExpOptimize.h
 *
 *  Created on: 20. 1. 2014
 *      Author: Tomas Pecka
 */

#ifndef REGEXPOPTIMIZE_H_
#define REGEXPOPTIMIZE_H_

#include <algorithm>
#include <functional>
#include <iterator>

#include <regexp/RegExp.h>

#include <regexp/unbounded/UnboundedRegExp.h>
#include <regexp/unbounded/UnboundedRegExpElements.h>

#include <regexp/formal/FormalRegExp.h>
#include <regexp/formal/FormalRegExpElements.h>

#include <exception/AlibException.h>

namespace regexp {

namespace simplify {

/*
 * Optimizes RegExp (or its subtree) using axioms defined in Melichar 2.87
 * (A1 to A10) and Melichar 2.95(V1 through V6 and V8, V9, V10)
 * All methods return new tree.
 *
 * List of optimization on nodes:
 *    - RegExpAlternation: A1, A2, A3, A4, A9, V2, V5, V6
 *    - RegExpConcatenation: A5, A6, A7, A8, V8, V9
 *    - RegExpIteration: A10, V1, V3, V4, V10
 *
 * Details: ( id : direction of optim. : optim )
 *  - A1 : -> : x + ( y + z ) = ( x + y ) + z = x + y + z
 *  - A2 : <- : x + y = y + x
 *  - A3 : -> : x + \0 = x
 *  - A4 : -> : x + x = x
 *  - A5 : -> : x(yz) = (xy)z = xyz
 *  - A6 : -> : \ex = x\e = x
 *  - A7 : -> : \0x = x\0 = \0
 *  - A8 : -> : x( y + z ) = xy + xz
 *  - A9 : -> : ( x + y )z = xz + yz
 *  - A10: <- : x* = \e + x*x
 *  - A11: <- : x* = ( \e + x )*
 *  - V1 : -> : \0* = \e
 *  - V2 : -> : x* + x = x*
 *  - V3 : -> : x** = x*
 *  - V4 : <- : ( x + y )* = (x*y*)*
 *  - V5 : <- : x*y = y + x*xy
 *  - V6 : <- : x*y = y + xx*y
 *  - V7 :    : bleh
 *  - V8 : -> : if \e in h(x) => xx* = x*
 *  - V9 : -> : (xy)*x = x(yx)*
 *  - V10: <- : ( x + y )* = ( x* + y* )*
 *
 *  - X1 : -> : a* + \e = a*
 */
class RegExpOptimize : public regexp::VisitableRegExpBase::const_visitor_type
{
public:
    static regexp::RegExp optimize( const regexp::RegExp & regexp );

    static regexp::UnboundedRegExp optimize( const regexp::UnboundedRegExp & regexp );
    static void optimize( regexp::UnboundedRegExpElement & regexp );

    static regexp::FormalRegExp optimize( const regexp::FormalRegExp & regexp );
    static void optimize( regexp::FormalRegExpElement & regexp );
private:
    regexp::FormalRegExpElement * optimize( regexp::FormalRegExpElement const * const & node ) const;

    regexp::UnboundedRegExpElement * optimize( regexp::UnboundedRegExpElement const * const & node ) const;
    regexp::UnboundedRegExpElement * optimize( regexp::UnboundedRegExpAlternation const * const & node ) const;
    regexp::UnboundedRegExpElement * optimize( regexp::UnboundedRegExpConcatenation const * const & node ) const;
    regexp::UnboundedRegExpElement * optimize( regexp::UnboundedRegExpIteration const * const & node ) const;
    regexp::UnboundedRegExpElement * optimize( regexp::UnboundedRegExpSymbol const * const & node ) const;
    regexp::UnboundedRegExpElement * optimize( regexp::UnboundedRegExpEpsilon const * const & node ) const;
    regexp::UnboundedRegExpElement * optimize( regexp::UnboundedRegExpEmpty const * const & node ) const;

    void Visit(void*, const regexp::UnboundedRegExp& regexp) const;
    void Visit(void*, const regexp::FormalRegExp& regexp) const;

private:
    bool A1( regexp::UnboundedRegExpAlternation * const & node ) const;
    bool A2( regexp::UnboundedRegExpAlternation * const & node ) const;
    bool A3( regexp::UnboundedRegExpAlternation * const & node ) const;
    bool A4( regexp::UnboundedRegExpAlternation * const & node ) const;
    bool A5( regexp::UnboundedRegExpConcatenation * const & node ) const;
    bool A6( regexp::UnboundedRegExpConcatenation * const & node ) const;
    bool A7( regexp::UnboundedRegExpConcatenation * const & node ) const;
    bool A8( regexp::UnboundedRegExpConcatenation * const & node ) const;
    bool A9( regexp::UnboundedRegExpConcatenation * const & node ) const;
    bool A10( regexp::UnboundedRegExpAlternation * const & node ) const;
    bool A11( regexp::UnboundedRegExpIteration * const & node ) const;
    bool V1( regexp::UnboundedRegExpIteration * const & node ) const;
    bool V2( regexp::UnboundedRegExpAlternation * const & node ) const;
    bool V3( regexp::UnboundedRegExpIteration * const & node ) const;
    bool V4( regexp::UnboundedRegExpIteration * const & node ) const;
    bool V5( regexp::UnboundedRegExpAlternation * const & node ) const;
    bool V6( regexp::UnboundedRegExpAlternation * const & node ) const;
    bool V8( regexp::UnboundedRegExpConcatenation * const & node ) const;
    bool V9( regexp::UnboundedRegExpConcatenation * const & node ) const;
    bool V10( regexp::UnboundedRegExpIteration * const & node ) const;

    bool X1( regexp::UnboundedRegExpAlternation * const & node ) const;

    bool S( regexp::FormalRegExpElement * & node ) const;
    bool A1( regexp::FormalRegExpElement * & node ) const;
    bool A2( regexp::FormalRegExpElement * & node ) const;
    bool A3( regexp::FormalRegExpElement * & node ) const;
    bool A4( regexp::FormalRegExpElement * & node ) const;
    bool A5( regexp::FormalRegExpElement * & node ) const;
    bool A6( regexp::FormalRegExpElement * & node ) const;
    bool A7( regexp::FormalRegExpElement * & node ) const;
    bool A8( regexp::FormalRegExpElement * & node ) const;
    bool A9( regexp::FormalRegExpElement * & node ) const;
    bool A10( regexp::FormalRegExpElement * & node ) const;
    bool A11( regexp::FormalRegExpElement * & node ) const;
    bool V1( regexp::FormalRegExpElement * & node ) const;
    bool V2( regexp::FormalRegExpElement * & node ) const;
    bool V3( regexp::FormalRegExpElement * & node ) const;
    bool V4( regexp::FormalRegExpElement * & node ) const;
    bool V5( regexp::FormalRegExpElement * & node ) const;
    bool V6( regexp::FormalRegExpElement * & node ) const;
    bool V8( regexp::FormalRegExpElement * & node ) const;
    bool V9( regexp::FormalRegExpElement * & node ) const;
    bool V10( regexp::FormalRegExpElement * & node ) const;

    bool X1( regexp::FormalRegExpElement * & node ) const;

    static const RegExpOptimize REG_EXP_OPTIMIZE;
};

} /* namespace simplify */

} /* namespace regexp */

#endif /* REGEXPNORMALIZE_H_ */
