/*
 * RegExpDerivation.cpp
 *
 *  Created on: 19. 1. 2014
 *      Author: Tomas Pecka
 */

#include "RegExpDerivation.h"

#include "../properties/RegExpEpsilon.h"

namespace regexp
{

regexp::RegExp RegExpDerivation::derivation(const regexp::RegExp& regexp, const string::LinearString& string)
{
    std::pair<regexp::RegExp*, const string::LinearString> out(nullptr, string);
    regexp.getData().Accept((void*) &out, RegExpDerivation::REGEXP_DERIVATION);
    regexp::RegExp res = std::move(*(out.first));
    delete out.first;
    return res;
}

regexp::FormalRegExp RegExpDerivation::derivation(const regexp::FormalRegExp& regexp, const string::LinearString& string)
{
    const regexp::FormalRegExpElement* newRegExp = regexp.getRegExp().clone();

    for(const auto& symbol : string.getContent())
    {
        std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> out(symbol, nullptr);
        newRegExp->Accept((void*) &out, RegExpDerivation::REGEXP_DERIVATION);
        delete newRegExp;
        newRegExp = out.second;
    }

    regexp::FormalRegExp res(std::move(*newRegExp));
    delete newRegExp;
    return res;
}

regexp::UnboundedRegExp RegExpDerivation::derivation(const regexp::UnboundedRegExp& regexp, const string::LinearString& string)
{
    const regexp::UnboundedRegExpElement* newRegExp = regexp.getRegExp().clone();

    for(const auto& symbol : string.getContent())
    {
        std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> out(symbol, nullptr);
        newRegExp->Accept((void*) &out, RegExpDerivation::REGEXP_DERIVATION);
        delete newRegExp;
        newRegExp = out.second;
    }

    regexp::UnboundedRegExp res(std::move(*newRegExp));
    delete newRegExp;
    return res;
}

// ----------------------------------------------------------------------------

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExp& regexp) const
{
    std::pair<regexp::RegExp*, const string::LinearString>& out = *((std::pair<regexp::RegExp*, const string::LinearString>*) userData);
    out.first = new regexp::RegExp(this->derivation(regexp, out.second));
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExp& regexp) const
{
    std::pair<regexp::RegExp*, const string::LinearString>& out = *((std::pair<regexp::RegExp*, const string::LinearString>*) userData);
    out.first = new regexp::RegExp(this->derivation(regexp, out.second));
}

// ----------------------------------------------------------------------------

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpAlternation& alternation) const
{
    std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);

    static_cast<const regexp::FormalRegExpElement&>(alternation.getLeftElement()).Accept(userData, *this);
    const regexp::FormalRegExpElement* leftDerivative = out.second;

    static_cast<const regexp::FormalRegExpElement&>(alternation.getRightElement()).Accept(userData, *this);
    const regexp::FormalRegExpElement* rightDerivative = out.second;

    out.second = new regexp::FormalRegExpAlternation(std::move(*leftDerivative), std::move(*rightDerivative));
    delete leftDerivative;
    delete rightDerivative;
}

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpConcatenation& concatenation) const
{
    std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);

    static_cast<const regexp::FormalRegExpElement&>(concatenation.getLeftElement()).Accept(userData, *this);
    const regexp::FormalRegExpElement* leftDerivative = out.second;
    const regexp::FormalRegExpElement* rightElement = concatenation.getRightElement().clone();
    out.second = new regexp::FormalRegExpConcatenation(std::move(*leftDerivative), std::move(*rightElement));
    delete leftDerivative;
    delete rightElement;

    if(regexp::properties::RegExpEpsilon::languageContainsEpsilon(concatenation.getLeftElement()))
    {
        const regexp::FormalRegExpElement *alternation = out.second, *rightDerivative = nullptr;
        static_cast<const regexp::FormalRegExpElement&>(concatenation.getRightElement()).Accept(userData, *this);
        rightDerivative = out.second;

        out.second = new regexp::FormalRegExpAlternation(std::move(*alternation), std::move(*rightDerivative));
        delete alternation;
        delete rightDerivative;
    }
}

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpIteration& iteration) const
{
    std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);

    static_cast<const regexp::FormalRegExpElement&>(iteration.getElement()).Accept(userData, *this);
    const regexp::FormalRegExpElement* iterDerivative = out.second;
    const regexp::FormalRegExpElement* iter = iteration.clone();

    out.second = new regexp::FormalRegExpConcatenation(std::move(*iterDerivative), std::move(*iter));
    delete iterDerivative;
    delete iter;
}

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpSymbol& symbol) const
{
    std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);

    if(out.first == symbol.getSymbol())
        out.second = new regexp::FormalRegExpEpsilon();
    else
        out.second = new regexp::FormalRegExpEmpty();
}


void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpEpsilon&) const
{
    std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);
    out.second = new regexp::FormalRegExpEmpty();
}

void RegExpDerivation::Visit(void* userData, const regexp::FormalRegExpEmpty&) const
{
    std::pair<alphabet::Symbol, regexp::FormalRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::FormalRegExpElement*>*) userData);
    out.second = new regexp::FormalRegExpEmpty();
}

// ----------------------------------------------------------------------------


void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpAlternation& alternation) const
{
    std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);
    regexp::UnboundedRegExpAlternation* ret = new regexp::UnboundedRegExpAlternation();

    for(const auto& child : alternation.getElements())
    {
        static_cast<const regexp::UnboundedRegExpElement&>(*child).Accept(userData, *this);
        ret->appendElement(std::move(*out.second));
        delete out.second;
    }

    out.second = ret;
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpConcatenation& concatenation) const
{
    std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);
    regexp::UnboundedRegExpAlternation* ret = new regexp::UnboundedRegExpAlternation();

    for(auto child = concatenation.getElements().begin(); child != concatenation.getElements().end(); child ++)
    {
        regexp::UnboundedRegExpConcatenation* concat = new regexp::UnboundedRegExpConcatenation();
        static_cast<const regexp::UnboundedRegExpElement&>(**child).Accept(userData, *this);
        concat->appendElement(std::move(*out.second));
        delete out.second;

        auto succeedingElement = child;
        while(++succeedingElement != concatenation.getElements().end())
        {
            regexp::UnboundedRegExpElement* tmp = (*succeedingElement)->clone();
            concat->appendElement(std::move(*tmp));
            delete tmp;
        }

        ret->appendElement(std::move(*concat));
        delete concat;

        if(regexp::properties::RegExpEpsilon::languageContainsEpsilon(**child))
            continue; // this IF construct is intentional "to match algorithm"
        break;
    }

    out.second = ret;
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpIteration& iteration) const
{
    std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);

    static_cast<const regexp::UnboundedRegExpElement&>(iteration.getElement()).Accept(userData, *this);
    const regexp::UnboundedRegExpElement* iterDerivative = out.second;
    const regexp::UnboundedRegExpElement* iter = iteration.clone();

    regexp::UnboundedRegExpConcatenation* ret = new regexp::UnboundedRegExpConcatenation();
    ret->appendElement(std::move(*iterDerivative));
    ret->appendElement(std::move(*iter));
    delete iterDerivative;
    delete iter;

    out.second = ret;
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpSymbol& symbol) const
{
    std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);

    if(out.first == symbol.getSymbol())
        out.second = new regexp::UnboundedRegExpEpsilon();
    else
        out.second = new regexp::UnboundedRegExpEmpty();
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpEpsilon&) const
{
    std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);
    out.second = new regexp::UnboundedRegExpEmpty();
}

void RegExpDerivation::Visit(void* userData, const regexp::UnboundedRegExpEmpty&) const
{
    std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*> &out = *((std::pair<alphabet::Symbol, regexp::UnboundedRegExpElement*>*) userData);
    out.second = new regexp::UnboundedRegExpEmpty();
}

const RegExpDerivation RegExpDerivation::REGEXP_DERIVATION;

} /* namespace regexp */
